"""Module which handles polytopes."""

import numpy as np
from typing import Optional
from scipy.optimize import linprog
import itertools


def inherently_nondominated(A: np.ndarray, epsilon: Optional[float] = 1e-06, method: Optional[str] = "highs") -> bool:
    """
    Check if a polytope is inherently nondominated:
    A polytope is inherently nondominated iff the polytope does not dominate itself.

    Args:
        A (np.ndarray): A polytope to be checked.
        epsilon (Optional[float], optional): precision parameter, see polytope_dominates for further details.
            Defaults to 1e-6.
        method (Optional[str], optional): Algorithm used to solve the optimization problems. Defaults to 'highs'.

    Returns:
        bool: is the given set inherently nondominated.
    """
    return not polytope_dominates(A, A, epsilon, method)


def polytope_dominates(
    k1: np.ndarray,
    k2: np.ndarray,
    epsilon: Optional[float] = 1e-6,
    method: Optional[str] = "highs"
) -> bool:
    """
    Check if polytope p(k1) dominates polytope p(k2) with epsilon certainty 
    by solving linear optimization problems [min_x c^T*x] using linprog from scipy.optimize.

    Args:
        k1 (np.ndarray): Corners of first polytope
        k2 (np.ndarray): Corners of seconds polytope
        epsilon (Optional[float], optional): precision parameter. Defaults to 1e-6
        method (Optional[str], optional): Algorithm used to solve the optimization problems.
            Defaults to 'highs'. See scipy.optimize.linprog for further details.

    Returns:
        bool: Does polytope p(k1) dominate polytope p(k2).
    """
    k1 = np.atleast_2d(k1)
    k2 = np.atleast_2d(k2)
    a, k = k1.shape
    b = k2.shape[0]

    # First optimization problem
    coef = np.hstack((np.ones(1), np.zeros(a+b)))
    lower_bounds = np.hstack((np.array([None]), np.zeros(a+b)))
    upper_bounds = np.hstack((np.array([None]), np.ones(a+b)))
    bounds = np.vstack((lower_bounds, upper_bounds)).T

    # Constructing the matrix A_ub in the constraint A_ub x <= b_ub
    A1 = np.hstack((-np.ones((k,1)), k1.T, -k2.T))
    A2 = np.hstack((np.zeros((1, 1)), np.ones((1,a)), np.zeros((1,b))))
    A3 = np.hstack((np.zeros((1, 1)), np.zeros((1,a)), np.ones((1,b))))
    A_ub = np.vstack((A1, np.zeros((2,a+b+1))))
    b_ub = np.zeros(k+2)

    A_eq = np.vstack((np.zeros((k,a+b+1)), A2, A3)) # Add k rows for correct size, will be ignored
    b_eq = np.hstack((np.zeros(k), np.ones(2))) 

    res = linprog(coef, A_ub, b_ub, A_eq, b_eq, bounds, method = method)

    if not res['success']: 
        print("unsuccessful optimization in first problem.")
    if res['fun'] < -epsilon: 
        return True

    if (abs(res['fun']) <= epsilon):
        lower_bounds = np.zeros(a+b)
        upper_bounds = np.ones(a+b)
        bounds = np.vstack((lower_bounds, upper_bounds)).T
        
        coef = np.hstack(([np.sum(k1, axis = 1), -np.sum(k2, axis = 1)]))
        A1 = np.hstack((np.ones((1,a)), np.zeros((1,b))))
        A2 = np.hstack((np.zeros((1,a)), np.ones((1,b))))
        A3 = np.hstack((k1.T, -k2.T))

        A_eq = np.vstack((A1, A2, np.zeros((k, a+b))))
        b_eq = np.hstack((np.ones(2), np.zeros(k)))

        A_ub = np.vstack((np.zeros((2, a+b)), A3))
        b_ub = np.zeros((k+2, 1))

        res = linprog(coef, A_ub, b_ub, A_eq, b_eq, bounds, method = method)
 
        if not res['success']: 
            print("unsuccessful optimization in second problem.")
        if res['fun'] < -epsilon: 
            return True
    return False


def generate_polytopes(simplices: np.ndarray) -> np.ndarray:
    """
    Generate polytopes from an array of indices which form simplices

    Args:
        arr (np.ndarray): An array of indices which form simplices. 
            In PAINT this is the array of simplices which form the Delaunay triangulation.

    Returns
        np.ndarray: An array of indices which form the polytopes
            that are generated from the given array. If a polytope has fewer
            outcomes than there are columns in the given array the first value of 
            the row representing the polytope is repeated until the lengths match.
    """
    simplices = np.sort(simplices)
    a,b = simplices.shape
    k = np.max(simplices) # point count
    F = (np.ones((b,k+1)) *np.arange(k+1)).T
    for i in range(a):
        for j in range(2,b+1):
            # All combinations of size j from row i
            addition = np.array(list(itertools.combinations(simplices[i], j)))
            # Add all combinations to F, so that we repeat the value at index 0 until enough values
            chunks = addition[:,0].shape[0] # How many chunks to split into
            repeated = np.split(np.repeat(addition[:,0], b-j), chunks) # Repeat the values at index 0
            addition = np.hstack((addition, repeated)) # Add the repeated values
            F = np.vstack((F, addition)) # Add new rows to F
    # F(F(:,1)==0,:) = []; ?
    return np.unique(F, axis = 0).astype(int)



"""This module implements methods for defining functions
which map vectors to vectors (of same or different lengths).

Examples include simple transformations such as translation or rotation,
and complicated mappings such as the creation of Preference Incorporated Spaces.
"""

__all__ = ["classificationPIS"]

from desdeo_tools.maps.preference_incorporated_space_RP import classificationPIS



from numba import njit
import numpy as np
import hvwfg as hv


@njit()
def epsilon_indicator(reference_front: np.ndarray, front: np.ndarray) -> float:
    """ Computes the additive epsilon-indicator between reference front and current approximating front.

    Args:
        reference_front (np.ndarray): The reference front that the current front is being compared to.
        Should be an one-dimensional array.
        front (np.ndarray): The front that is compared. Should be one-dimensional array with the same shape as
        reference_front.

    Returns:
        float: The factor by which the approximating front is worse than the reference front with respect to all
        objectives.
    """
    eps = 0.0
    for i in range(reference_front.size):
        value = front[i] - reference_front[i]
        if value > eps:
            eps = value
    return eps


@njit()
def epsilon_indicator_ndims(reference_front: np.ndarray, front: np.ndarray) -> float:
    """ Computes the additive epsilon-indicator between reference front and current approximating front.
    Args:
        reference_front (np.ndarray): The reference front that the current front is being compared to.
        Should be set of arrays, where the rows are the solutions and the columns are the objective dimensions.
        front (np.ndarray): The front that is compared. Should be one-dimensional array.
    Returns:
        float: The factor by which the approximating front is worse than the reference front with respect to all
        objectives.
    """

    eps = 0.0
    ref_len = reference_front.shape[0]
    front_len = front.shape[0]
    value = 0

    for i in np.arange(ref_len):
        for j in np.arange(front_len):
            value = front[j] - reference_front[i][j]
            if value > eps:
                eps = value

    return eps


def hypervolume_indicator(reference_front: np.ndarray, front: np.ndarray) -> float:
    """ Computes the hypervolume-indicator between reference front and current approximating point.

    Args:
        reference_front (np.ndarray): The reference front that the current front is being compared to.
        Should be set of arrays, where the rows are the solutions and the columns are the objective dimensions.
        front (np.ndarray): The front that is compared. Should be 2D array.

    Returns:
        float: Measures the volume of the objective space dominated by an approximation set.
    """
    return hv.wfg(reference_front, front.reshape(-1))


if __name__ == "__main__":
    x = np.array([[1, 0], [0.5, 0.5], [0, 1], [1.5, 0.75]])
    ref = np.array([[2.0, 2.0]])
    print(epsilon_indicator(x, ref))
    print(hypervolume_indicator(x, ref))

    x_simple = np.array([[0.0, 0.0]])
    ref_simple = np.array([[2.0, 1.0]])
    print(epsilon_indicator(x_simple, ref_simple))

    obj = np.array([[0.3, 0.6, 1.0], [0.4, 0.4, 1.2], [0.6, 0.2, 0.3]])

    print(epsilon_indicator(obj, ref))
    ref_hv = np.array([[1.1, 1.1, 1.1]])
    print(hypervolume_indicator(obj, ref_hv))
    ref_hv2 = np.array([[2.0, 2.0, 2.0]])
    print(hypervolume_indicator(obj, ref_hv2))

    print("\n========= PERFORMANCE TEST ===========")
    objvalues = 1000
    objdims = 77
    solpoints = np.random.randint(1, objdims)
    print(solpoints)

    x_hard = np.array(np.random.rand(objvalues, objdims))
    # print(x_hard)
    ref_hard = np.array(np.random.rand(solpoints, objdims))
    # print(ref_hard)
    ref_hard_vector = np.array(np.random.rand(1, objdims))
    print(epsilon_indicator(x_hard, ref_hard))
    print(hypervolume_indicator(x_hard, ref_hard_vector))



"""
This module implements various small tools that may be useful during optimization using
either the MCDM or the EA methods.
"""

__all__ = [
    "dominates",
    "non_dominated",
    "fast_non_dominated_sort",
    "fast_non_dominated_sort_indices",
    "classification_to_reference_point",
    "epsilon_indicator",
    "epsilon_indicator_ndims",
    "hypervolume_indicator",
    "polytope_dominates",
    "generate_polytopes",
    "inherently_nondominated",
]

from desdeo_tools.utilities.polytopes import (
    polytope_dominates,
    generate_polytopes,
    inherently_nondominated,
)

from desdeo_tools.utilities.quality_indicator import (
    epsilon_indicator,
    epsilon_indicator_ndims,
    hypervolume_indicator,
)

from desdeo_tools.utilities.fast_non_dominated_sorting import (
    dominates,
    non_dominated,
    fast_non_dominated_sort,
    fast_non_dominated_sort_indices,
)

from desdeo_tools.utilities.preference_converters import (
    classification_to_reference_point,
)



from desdeo_tools.scalarization.GLIDE_II import (
    GLIDEBase,
    NIMBUS_GLIDE,
    reference_point_method_GLIDE,
    STOM_GLIDE,
)
from desdeo_tools.utilities import classification_to_reference_point
import numpy as np
from typing import Type, List, Union, Dict


class PreferenceIncorporatedSpaceError(Exception):
    "Raised when an error related to the preference incorporated space is encountered."


class PreferenceIncorporatedSpace:
    def __init__(
        self,
        scalarizers: List[Type[GLIDEBase]],
        utopian: np.ndarray,
        nadir: np.ndarray,
        preference: dict,
        rho: float = 1e-6,
    ):
        self.scalarizers = scalarizers
        self.update_map(utopian=utopian, nadir=nadir, rho=rho)

    def update_map(
        self,
        utopian: np.ndarray,
        nadir: np.ndarray,
        scalarizers: List[Type[GLIDEBase]] = None,
        rho: float = 1e-6,
    ):
        if scalarizers is None:
            scalarizers = self.scalarizers
        else:
            self.scalarizers = scalarizers

        if utopian is not None:
            self.utopian = utopian
        if nadir is not None:
            self.nadir = nadir

        self.initialized_scalarizers = [
            scalarizer(utopian=utopian, nadir=nadir, rho=rho)
            for scalarizer in scalarizers
        ]

        self.preferences = []
        self.has_additional_constraints = False
        self.constrained_scalarizers = []

        for scalarizer in self.initialized_scalarizers:
            # self.required_keys = scalarizer.required_keys.keys()

            self.constrained_scalarizers.append(scalarizer.has_additional_constraints)
            self.has_additional_constraints = (
                self.has_additional_constraints or scalarizer.has_additional_constraints
            )

    def __call__(self, objective_vector: np.ndarray):
        mapped_vectors = np.zeros(
            (len(objective_vector), len(self.initialized_scalarizers))
        )
        for i, scalarizer in enumerate(self.initialized_scalarizers):
            mapped_vectors[:, i] = scalarizer(objective_vector, self.preferences[i])
        return mapped_vectors

    def evaluate_constraints(self, objective_vector: np.ndarray):
        if not self.has_additional_constraints:
            return None
        constraints = None
        for i, (scalarizer, has_constraints) in enumerate(
            zip(self.initialized_scalarizers, self.constrained_scalarizers)
        ):
            if not has_constraints:
                continue
            constraints = np.hstack(
                (
                    constraints,
                    scalarizer.evaluate_constraints(
                        objective_vector, self.preferences[i]
                    ),
                )
            )


class classificationPIS:
    """Implements the preference incorporated space mapping which uses the classification preference.

    Args:
        scalarizers (List[Type[GLIDEBase]]): Scalarizers to be used to create the PIS.
            Should include atleast one scalarizer. NIMBUS should not be included as it is added automatically.
        utopian (np.ndarray): The utopian point of the problem.
        nadir (np.ndarray): The nadir point of the problem.
        rho (float, optional): The augmentation factor used in the different scalarizers. Defaults to 1e-6.
    """

    def __init__(
        self,
        scalarizers: List[Type[GLIDEBase]],
        utopian: np.ndarray,
        nadir: np.ndarray,
        rho: float = 1e-6,
    ):
        self.nimbus: Union[NIMBUS_GLIDE, None] = None
        self.nimbus_copycat: Union[reference_point_method_GLIDE, None] = None
        self.scalarizers = scalarizers
        self.update_map(utopian=utopian, nadir=nadir, rho=rho)

    def update_map(
        self,
        utopian: np.ndarray,
        nadir: np.ndarray,
        scalarizers: List[Type[GLIDEBase]] = None,
        rho: float = 1e-6,
    ):
        if scalarizers is None:
            scalarizers = self.scalarizers
        else:
            self.scalarizers = scalarizers

        if utopian is not None:
            self.utopian = utopian
        if nadir is not None:
            self.nadir = nadir

        self.nimbus = NIMBUS_GLIDE(utopian=utopian, nadir=nadir)
        self.nimbus_copycat = reference_point_method_GLIDE(utopian=utopian, nadir=nadir)

        self.initialized_scalarizers = [
            scalarizer(utopian=utopian, nadir=nadir, rho=rho)
            for scalarizer in scalarizers
        ]
        self.has_additional_constraints = False

    def update_preference(self, preference: dict):

        self.preference = preference
        if "classifications" in preference.keys():
            self.classification_preference = preference
            self.RP_preference = classification_to_reference_point(
                preference, ideal=self.utopian, nadir=self.nadir
            )
        else:
            raise PreferenceIncorporatedSpaceError(
                "Classification preference expected."
            )

    def __call__(self, objective_vector: np.ndarray):

        # IOPIS/NIMBUS logic
        nimbus_obj = self.nimbus(
            objective_vector=objective_vector, preference=self.classification_preference
        )
        nimbus_constraint = self.nimbus.evaluate_constraints(
            objective_vector, self.classification_preference
        )
        feasible = np.all(nimbus_constraint > 0, axis=1)

        if not feasible.any():
            nimbus_optimal = objective_vector[nimbus_constraint.max(axis=1).argmax()]
        else:
            nimbus_obj[~feasible] = np.inf
            nimbus_optimal = objective_vector[nimbus_obj.argmin()]

        # IOPIS mapping
        mapped_vectors = np.zeros(
            (len(objective_vector), len(self.initialized_scalarizers) + 1)
        )

        mapped_vectors[:, 0] = self.nimbus_copycat(
            objective_vector=objective_vector,
            preference={"reference point": nimbus_optimal},
        )

        for i, scalarizer in enumerate(self.initialized_scalarizers):
            mapped_vectors[:, i + 1] = scalarizer(objective_vector, self.RP_preference)
        return mapped_vectors


class MultiDMPIS(PreferenceIncorporatedSpace):
    def __init__(
        self,
        utopian: np.ndarray,
        num_DM: int = 2,
        scalarizer: Type[GLIDEBase] = STOM_GLIDE,
        nadir: np.ndarray = None,
        rho: float = 1e-6,
    ):
        super().__init__([scalarizer] * num_DM, utopian, nadir, rho)

    def update_preference(self, preference: List[Dict]):
        self.preferences = preference
